/*
 *  Raster graphics library
 *
 *  Copyright (c) 1997-2002 Alfredo K. Kojima
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef WRASTERP_H_
#define WRASTERP_H_

#include <config.h>


#include "x11/wraster.h"


#ifdef HAVE_HERMES

# include <Hermes/Hermes.h>

typedef struct RHermesData {
    HermesHandle palette;
    HermesHandle converter;
} RHermesData;

#endif


#endif
