/* 
   externs.m

   External data

   Copyright (C) 1997 Free Software Foundation, Inc.

   Author:  Adam Fedor <fedor@gnu.org>
   Date: Dec 1998
   
   This file is part of the GNUstep GUI Library.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; see the file COPYING.LIB.
   If not, see <http://www.gnu.org/licenses/> or write to the 
   Free Software Foundation, 51 Franklin Street, Fifth Floor, 
   Boston, MA 02110-1301, USA.
*/ 

#include <Foundation/NSString.h>

NSString *DPSconfigurationerror = @"DPSconfigurationerror: %@";
NSString *DPSinvalidaccess = @"DPSinvalidaccess: %@";
NSString *DPSinvalidcontext = @"DPSinvalidcontext: %@";
NSString *DPSinvalidexit = @"DPSinvalidexit: %@";
NSString *DPSinvalidfileaccess = @"DPSinvalidfileaccess: %@";
NSString *DPSinvalidfont = @"DPSinvalidfont: %@";
NSString *DPSinvalidid = @"DPSinvalidid: %@";
NSString *DPSinvalidrestore = @"DPSinvalidrestore: %@";
NSString *DPSinvalidparam = @"DPSinvalidparam: %@";
NSString *DPSioerror = @"DPSioerror: %@";
NSString *DPSlimitcheck = @"DPSlimitcheck: %@";
NSString *DPSnocurrentpoint = @"DPSnocurrentpoint: %@";
NSString *DPSnulloutput = @"DPSnulloutput: %@";
NSString *DPSrangecheck = @"DPSrangecheck: %@";
NSString *DPSstackoverflow = @"DPSstackoverflow: %@";
NSString *DPSstackunderflow = @"DPSstackunderflow: %@";
NSString *DPStypecheck = @"DPStypecheck: %@";
NSString *DPSundefined = @"DPSundefined: %@";
NSString *DPSundefinedfilename = @"DPSundefinedfilename: %@";
NSString *DPSundefinedresource = @"DPSundefinedresource: %@";
NSString *DPSundefinedresult = @"DPSundefinedresult: %@";
NSString *DPSunmatchedmark = @"DPSunmatchedmark: %@";
NSString *DPSunregistered = @"DPSunregistered: %@";
NSString *DPSVMerror = @"DPSVMerror: %@";
